/*{{{  revision history*/
/*
2003-11-16  V 0.1  First draft
2003-11-25  V 1.00 First tested version
2003-11-26  V 1.01 Removed call to PhaseFudge in DecompFFT as this clearly
                   does not work correctly on the SpinSynth data.
            V 1.03 Permute axes for SR2 so Z is spin axis.
2003-12-05  V 1.04 Modified ReadCalData to ignore extra columns in cal file.
                   Implemented -s and -e on DSD files.
            V 1.05
2004-02-27  V 1.06 Check packet sizes.
                   SR2 rotation default set to 107.25.
                   New -a option to supress Reset Pulse Spike.
                   SRP times from Correlator data supports >1 acf/spin.
				   Time tag processing extrapolates before start of HK file.
				   Remaining 'decom' errors suppressed unless -v 16.
				   New -l option to suppress spikes based on power limit.
/*
/*}}}  */

/*{{{  include files*/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#include <string.h>
#include <math.h>

#include "tlmproc.h"
#include "fftl2.h"
/*}}}  */

/*{{{  structures*/
struct dcomplex {
    double x,y;
};

struct caldata {
    double Sx;
    struct dcomplex Syz, Dyz;
    };

struct specmat {
    double Px, Py, Pz;
    struct dcomplex Cxy, Cyz, Cxz;
    };
/*}}}  */

/*{{{  global variables*/
/*{{{  static int LRStartBin[] = {  1,  2,  3,  4,  5,  6,   7,    8,   9,*/
static int LRStartBin[] = {  1,  2,  3,  4,  5,  6,   7,    8,   9,
                            10, 11, 13, 15, 18,  21,  25,  30,  35,
                            42, 50, 59, 71, 84, 100, 119, 141, 168, 200 };

static int LRNumBins[] = {  1,  1,  1,  1,  1,  1,  1,  1,  1,
                         1,  2,  2,  3,  3,  4,  5,  5,  7,
                         8,  9, 12, 13, 16, 19, 22, 27, 32 };
/*}}}  */

/* Pointers to strings for command line options, NULL = not specified */
char *STA_HK_name = NULL, *SC_HK_name = NULL, *STA_SCI_name = NULL,
    *Cal_name = NULL, *DSD_name = NULL, *Out_name = NULL;

FILE *SC_HK_file = NULL, *Out_file = NULL;
FILE *HK_file = NULL, *SCI_file = NULL, *DSD_file = NULL;

struct caldata CalibrationData[27];
int CalVersion;

/* Spin phase variables */
double SunRefPulse, MeasuredSpinPeriod, NominalSpinPeriod = 4.0;
double CorrelSRPOffset;
int CorrelBitRateIndex;
int ValidSpin;

double SR2Rotation = 107.25;  /* Amount to rotate spectral matrix from despun
                                 antenna co-ordinates to get SR2 co-ordinates. */
double PowerLimit = 1.0e10;

long StartTime = 0;
long EndTime = 86400;

int RstSpikeMin = 0;       /* Minimum MC count for reset pulse spike */
int RstSpikeMax = 0;       /* Maximum MC count for reset pulse spike */
int BugPhaseMin = 2038;    /* Minimum spin phase for de-spin bug */
int BugPhaseMax = 2048;    /* Maximum spin phase for de-spin bug */

int TransformControl = -2; /* Control of co-ordinate transforms:
                           -3  de-spin off : antenna co-ords rotated
                               de-spin on  : de-spun antenna co-ords
                           -2  de-spin off : antenna co-ords
                               de-spin on  : SR2 co-ords
                           -1  no change
                            0  antenna co-ords
                            1  de-spun antenna co-ords
                            2  SR2 co-ords. */

int Verbosity;

int OutputCount = 0, SkipCount = 0;

int GroundStation = 0;
int TEDDiagnostic = 0;

/*}}}  */

/* Utility functions */
/*{{{  interpol (xval, lookup, looksize, yval)*/
interpol (xval, lookup, looksize, yval)
double xval, lookup[][2], *yval;
int looksize;
/* Function to do linear interpolation. Input value is 'xval',
   output value is '*yval', function value is 0 if OK, -1 if
   'xval' too negative, +1 if too positive. 'lookup' is two
   dimensional array [looksize][2] containing look up table with
   x values [][1], and y values [][0].                           */
{
int index;

if (xval < lookup[0][1]) return (-1);

index = 1;
while (xval > lookup[index][1]) {
  index++;
  if (index == looksize) return (1);
  }

*yval = (lookup[index-1][0] * (lookup[index][1] - xval) +
          lookup[index][0] * (xval - lookup[index-1][1])) /
         (lookup[index][1] - lookup[index-1][1]);

return (0);
}
/*}}}  */
/*{{{  double CCSDS_time_ext (char *time_str)*/
double CCSDS_time_ext (char *time_str)
{
/* Encodes extended CCSDS time string (including milli-seconds) as a double
   giving number of seconds from 1970-01-01 00:00:00.000 */
struct tm timestruc;
double seconds;

if (sscanf (time_str, "%d-%d-%dT%d:%d:%lf",
            &timestruc.tm_year, &timestruc.tm_mon, &timestruc.tm_mday,
            &timestruc.tm_hour, &timestruc.tm_min, &seconds) != 6)
            return (0.0);

timestruc.tm_year -= 1900; /* Years measured from 1900 */
timestruc.tm_mon  -= 1;    /* Months start at zero */
timestruc.tm_isdst = 0;    /* No daylight saving */
timestruc.tm_sec   = 0;    /* Seconds are added later */

return ((double) mktime (&timestruc) + seconds);
}
/*}}}  */

/*{{{  int PsLog (double x)*/
int PsLog (double x)
{
    int result;

    if (x <= 0.0) return 0;

    result = 0;

    while (x >= 2.0) {
        x *= 0.5;
        result += 8;
    }

    result += (int)((x - 1.0) * 8.0);

    return result;
}

/*}}}  */
/*{{{  double InvPsLog (int x)*/
double InvPsLog (int x)
{
    double result;

    result = (x % 8) / 8.0 + 1.0;

    while (x >= 8) {
        x -= 8;
        result *= 2.0;
    }

    return result;
}

/*}}}  */
/*{{{  double V2pHz (int comp, int chan, unsigned char *pdata)*/
double V2pHz (int comp, int chan, unsigned char *pdata)
/* Returns power spectral density in V squared per Hz */
{
    double power;
    int offset, count, logenergy;

    /*{{{  get offset of required item*/
    offset = chan + 6;
    switch (comp) {
        case spHR:
        break;
        case spXP:  /* X */
        if (disableX(pdata)) return 0.0;
        break;
        case spYP:  /* Y */
        if (disableY(pdata)) return 0.0;
        if (!disableX(pdata)) offset += 27;
        break;
        case spZP:  /* Z */
        if (disableZ(pdata)) return 0.0;
        if (!disableY(pdata)) offset += 27;
        if (!disableX(pdata)) offset += 27;
        break;
        default:
        return 0.0;
    }
    /*}}}  */

    count = fftcount (pdata);
    logenergy = pdata[offset] + PsLog (count);
    power = (InvPsLog (logenergy) + InvPsLog (logenergy + 1)) /  (count * 2);
    if (higain (pdata)) power /= 16.0;
    if (comp != spHR) power /= LRNumBins[chan];
    return power / 3.5791e09;
}

/*}}}  */
/*{{{  double PhaseRad (int comp, int chan, unsigned char *pdata)*/
double PhaseRad (int comp, int chan, unsigned char *pdata)
/* Returns phase in radians for select component and channel */
{
    int offset, code, quad;
    double phase;

    /*{{{  determine where the phase information is located*/
    offset = chan + 27 * 3 + 6;
    switch (comp) {
        case 0:  /* XY */
        if (disableX(pdata) || disableY(pdata)) return 0.0;
        if (disableZ(pdata)) offset -= 27;
        break;
        case 1:  /* YZ */
        if (disableY(pdata) || disableZ(pdata)) return 0.0;
        if (disableX(pdata)) offset -= 27;
        else offset += 27;
        break;
        case 2:  /* ZX */
        if (disableZ(pdata) || disableX(pdata)) return 0.0;
        if (disableY(pdata)) offset -= 27;
        else offset += 27 * 2;
        break;
        default:
        return 0.0;
    }
    /*}}}  */

    code = pdata[offset];
    quad = code / 64;
    code = code % 64;

    if (code > 32) phase = 3.14159/2.0 - atan ((64 - code) / 32.0);
    else phase = atan (code / 32.0);

    phase = phase + quad * 3.14159/2.0;

    return phase;
}
/*}}}  */
/*{{{  double PhaseFudge (int chan, unsigned char *pdata)*/
double PhaseFudge (int chan, unsigned char *pdata)
{
    double phasesum;
    double totalcount;
    double snr2;
    double fudge;

    phasesum = fabs (PhaseRad (0, chan, pdata) +
                     PhaseRad (1, chan, pdata) +
                     PhaseRad (2, chan, pdata));

    totalcount = LRNumBins[chan] * fftcount (pdata);

    if (phasesum < 0.01) fudge = 1.0;
    else if (phasesum > 1.56) fudge = sqrt (1.0 / totalcount);
    else {
       snr2 = 1.0 / (tan (phasesum) * sqrt ((double) totalcount));
       fudge = sqrt (snr2 * snr2 + 1.0 / totalcount) / (snr2 + 1.0);
    }
    return fudge;
}
/*}}}  */
/*{{{  int ReadDSDPkt(FILE *fp, int bufsize, unsigned char *buf, double *scet)*/
int ReadDSDPkt(FILE *fp, int bufsize, unsigned char *buf, double *scet)
{
    struct tm timestruc;
    unsigned char dsd_header[64];
    unsigned int size, source, millis;

    do {
        /* Read DSD header as array of bytes */
        if (fread (dsd_header, sizeof dsd_header, 1, fp) < 1) return -1;

        /* Extract and check packet size */
        size   = (dsd_header[12] * 0x100 + dsd_header[13]) * 2;
        if (size + 4 > bufsize) return -2;

    } while (size == 0);

    /* Extract and check source instrument (FFT = 1) */
    source   = dsd_header[8] * 0x100 + dsd_header[9];
    if (source != 1) return -3;

	/* Extract ground station and diagnostics words */
	GroundStation = dsd_header[6]  * 0x100 + dsd_header[7];
	TEDDiagnostic = dsd_header[10] * 0x100 + dsd_header[11];

    /* Extract and convert SPRT */
    timestruc.tm_year = dsd_header[15];
    timestruc.tm_mon  = dsd_header[17] - 1;
    timestruc.tm_mday = dsd_header[19];
    timestruc.tm_hour = dsd_header[21];
    timestruc.tm_min  = dsd_header[23];
    timestruc.tm_sec  = dsd_header[25];
    millis = dsd_header[26] * 0x100 + dsd_header[27];

    timestruc.tm_isdst = 0;
    *scet = mktime (&timestruc) + millis * 0.001;

    /* Fill in the packet descriptor and time tag bytes */
    buf[0] = (source << 3) + (dsd_header[33] << 1); /* source + UDEF bits */
    buf[1] = 0;                   /* not used */
    buf[2] = dsd_header[30];      /* msb time tag */
    buf[3] = dsd_header[31];      /* lsb time tag */

    /* Read the packet into the rest of the buffer */
    if (fread (buf + 4, size, 1, fp) < 1) return -4;

    /* Return the packet size */
    return size + 4;

}
/*}}}  */
/*{{{  int DecompFFT (unsigned char *pdata, double pkt_time, struct specmat *pSM)*/
int DecompFFT (unsigned char *pdata, double pkt_time, struct specmat *pSM)
{
    int chan;
    double magnitude, phase;
    struct dcomplex Sy, Sz;

    if (udefbits (pdata) != 1 && udefbits (pdata) != 3) return 0;

    for (chan = 0; chan < 27; chan ++) {
        /* Calculate y and z transfer functions taking account of
           spin phase and de-spin mode */
        /*{{{  */
        if (despinoff (pdata)) {
            Sy.x = CalibrationData[chan].Syz.x + CalibrationData[chan].Dyz.x;
            Sy.y = CalibrationData[chan].Syz.y + CalibrationData[chan].Dyz.y;
            Sz.x = CalibrationData[chan].Syz.x - CalibrationData[chan].Dyz.x;
            Sz.y = CalibrationData[chan].Syz.y - CalibrationData[chan].Dyz.y;
        }
        else {
        /* For now we use only the average transfer function of y and z. */
            Sy.x = CalibrationData[chan].Syz.x;
            Sy.y = CalibrationData[chan].Syz.y;
            Sz.x = CalibrationData[chan].Syz.x;
            Sz.y = CalibrationData[chan].Syz.y;
        }
        /*}}}  */

        /*{{{  power spectra*/
        /* Power spectrum X component */
        /* CalibrationData[].Sx gives (real) transfer function of antenna in V/nT,
           so to convert V squared to nT squared, we divide by Sx squared. */
        pSM[chan].Px = V2pHz (spXP, chan, pdata) / (CalibrationData[chan].Sx * CalibrationData[chan].Sx);
        
        /* Power spectra Y component */
        pSM[chan].Py = V2pHz (spYP, chan, pdata) / (Sy.x * Sy.x + Sy.y * Sy.y);
        
        /* Power spectra Z component */
        pSM[chan].Pz = V2pHz (spZP, chan, pdata) / (Sz.x * Sz.x + Sz.y * Sz.y);
        /*}}}  */

        /* Cross spectrum, Cxy */
        /* As we don't have the magnitude of the cross spectra,
           for now we use the square root of the product of the
           corresponding power spectra. */
        magnitude = sqrt (pSM[chan].Px * pSM[chan].Py);
        phase = -PhaseRad (0, chan, pdata) + atan2 (Sy.y, Sy.x);
        pSM[chan].Cxy.x = magnitude * cos (phase);
        pSM[chan].Cxy.y = magnitude * sin (phase);

        /* Cross spectrum, Cyz */
        magnitude = sqrt (pSM[chan].Py * pSM[chan].Pz);
        phase = -PhaseRad (1, chan, pdata) + atan2 (Sz.y, Sz.x) - atan2 (Sy.y, Sy.x);
        pSM[chan].Cyz.x = magnitude * cos (phase);
        pSM[chan].Cyz.y = magnitude * sin (phase);

        /* Cross spectrum, Czx -> Cxz */
        magnitude = sqrt (pSM[chan].Pz * pSM[chan].Px);
        phase = PhaseRad (2, chan, pdata) + atan2 (Sz.y, Sz.x);
        pSM[chan].Cxz.x = magnitude * cos (phase);
        pSM[chan].Cxz.y = magnitude * sin (phase);
    }
    return 1;
}

/*}}}  */
/*{{{  int RotateX (struct specmat *pSM, double theta)*/
int RotateX (struct specmat *pSM, double theta)
/* Transforms the spectral matrix by rotating the axes through an angle theta
   about the X axis measured from +Y towards +Z. */
{
   struct specmat input[1];
   double costheta, sintheta;

   memcpy (input, pSM, sizeof input);
   costheta = cos (theta);
   sintheta = sin (theta);

   /* Px is unchanged by rotation */

   pSM->Py    = input->Py    * costheta * costheta +
                input->Cyz.x * costheta * sintheta * 2 +
                input->Pz    * sintheta * sintheta;

   pSM->Pz    = input->Pz    * costheta * costheta -
                input->Cyz.x * costheta * sintheta * 2 +
                input->Py    * sintheta * sintheta;

   pSM->Cyz.x = input->Cyz.x * costheta * costheta +
                (input->Pz - input->Py) * costheta * sintheta -
                input->Cyz.x * sintheta * sintheta;

   /* Imaginary part of Cyz is unchanged by rotation */

   pSM->Cxy.x = input->Cxy.x * costheta + input->Cxz.x * sintheta;
   pSM->Cxy.y = input->Cxy.y * costheta + input->Cxz.y * sintheta;

   pSM->Cxz.x = input->Cxz.x * costheta - input->Cxy.x * sintheta;
   pSM->Cxz.y = input->Cxz.y * costheta - input->Cxy.y * sintheta;

   return 1;
}
/*}}}  */
/*{{{  int RotateY (struct specmat *pSM, double theta)*/
int RotateY (struct specmat *pSM, double theta)
/* Transforms the spectral matrix by rotating the axes through an angle theta
   about the Y axis measured from +Z towards +X. */
{
   struct specmat input[1];
   double costheta, sintheta;

   memcpy (input, pSM, sizeof input);
   costheta = cos (theta);
   sintheta = sin (theta);

   /* Py is unchanged by rotation */

   pSM->Pz    = input->Pz    * costheta * costheta +
                input->Cxz.x * costheta * sintheta * 2 +
                input->Px    * sintheta * sintheta;

   pSM->Px    = input->Px    * costheta * costheta -
                input->Cxz.x * costheta * sintheta * 2 +
                input->Pz    * sintheta * sintheta;

   pSM->Cxz.x = input->Cxz.x * costheta * costheta +
                (input->Px - input->Pz) * costheta * sintheta -
                input->Cxz.x * sintheta * sintheta;

   /* Imaginary part of Czx is unchanged by rotation */

   pSM->Cyz.x = input->Cyz.x * costheta + input->Cxy.x * sintheta;
   pSM->Cyz.y = input->Cyz.y * costheta - input->Cxy.y * sintheta;

   pSM->Cxy.x = input->Cxy.x * costheta - input->Cyz.x * sintheta;
   pSM->Cxy.y = input->Cxy.y * costheta + input->Cyz.y * sintheta;

   return 1;
}
/*}}}  */
/*{{{  int ListData (double cen_time, struct specmat *pSM)*/
int ListData (double cen_time, int coords, struct specmat *pSM)
{
    int i;
    struct specmat avdata;

    memset (&avdata, 0, sizeof avdata);
    for (i = 0; i < 27; i++) {
        avdata.Px    += pSM[i].Px    * LRNumBins[i] * 20.0;
        avdata.Py    += pSM[i].Py    * LRNumBins[i] * 20.0;
        avdata.Pz    += pSM[i].Pz    * LRNumBins[i] * 20.0;
        avdata.Cxy.x += pSM[i].Cxy.x * LRNumBins[i] * 20.0;
        avdata.Cxy.y += pSM[i].Cxy.y * LRNumBins[i] * 20.0;
        avdata.Cyz.x += pSM[i].Cyz.x * LRNumBins[i] * 20.0;
        avdata.Cyz.y += pSM[i].Cyz.y * LRNumBins[i] * 20.0;
        avdata.Cxz.x += pSM[i].Cxz.x * LRNumBins[i] * 20.0;
        avdata.Cxz.y += pSM[i].Cxz.y * LRNumBins[i] * 20.0;
    }

    printf ("%s %1d %10.2le %10.2le %10.2le %10.2le,%10.2le %10.2le,%10.2le %10.2le,%10.2le\n",
            doubledate (cen_time), coords,
            avdata.Px, avdata.Py, avdata.Pz,
            avdata.Cxy.x, avdata.Cxy.y,
            avdata.Cyz.x, avdata.Cyz.y,
            avdata.Cxz.x, avdata.Cxz.y);

    return 1;
}
/*}}}  */
/*{{{  int DespinBug (unsigned char *pdata, double pkt_time)*/
int DespinBug (unsigned char *pdata, double pkt_time)
/* Returns 1 if the data packet is likely to be effected by the de-spin bug.
   Returns 0 if data OK, or no spin phase information available. */
{
    /* This table gives the offsets of each raw time series back from the
       packet timetag. Only valid when all components enabled. */
    static double RawTSOffsets[] = { 0.0000, 0.1911, 0.3933, 0.5956, 0.7978 };

    double rawTStimetag;
    int rawTSindex;
    int spinphase;

    /* Can't check for bug if no spin phase info */
    if (!ValidSpin) return 0;

    /* No bug if de-spin off */
    if (despinoff (pdata)) return 0;

    /* If any component disabled, different check required */
    if (disableXYZ (pdata)) return 0;

    for (rawTSindex = 0; rawTSindex < fftcount (pdata) && rawTSindex < 5; rawTSindex++) {
        rawTStimetag = pkt_time - RawTSOffsets[rawTSindex];
        spinphase = (int)(4096.0 * (rawTStimetag - SunRefPulse) /
                          MeasuredSpinPeriod) & 4095;
        if ((spinphase >= BugPhaseMin) && (spinphase <= BugPhaseMax)) return spinphase;
        /* For V 1.06 onwards, we also handle the case where we want to exclude a
           region around zero phase, specified by setting min > max. */
        if ((BugPhaseMin >  BugPhaseMax) && (spinphase <= BugPhaseMax)) return spinphase;
        if ((BugPhaseMin >  BugPhaseMax) && (spinphase >= BugPhaseMin)) return spinphase;
    }

    return 0;
}

/*}}}  */

/*{{{  int ResetSpike (unsigned char *pdata)*/
int ResetSpike (unsigned char *pdata)
/* Returns 1 if the data packet is likely to be effected by the reset pulse spike.
   Returns 0 if data OK. */
{
    /* This table gives the offsets in MC counts of each raw time series relative to
       packet MC counts. Only valid when all components enabled. */
    static int RawTSMCoffset[] = { 0, 172, 354, 536, 718 };

    int rawTSMCcount;
    int rawTSindex;

    /* If any component disabled, different check required */
    if (disableXYZ (pdata)) return 0;

    for (rawTSindex = 0; rawTSindex < fftcount (pdata); rawTSindex++) {
        rawTSMCcount = samp_clks (pdata) - RawTSMCoffset[rawTSindex];
        if (rawTSMCcount < 0) rawTSMCcount += 4637;
        if ((rawTSMCcount >= RstSpikeMin) && (rawTSMCcount <= RstSpikeMax)) return rawTSMCcount;
        if ((RstSpikeMin >  RstSpikeMax) && (rawTSMCcount <= RstSpikeMax)) return rawTSMCcount;
        if ((RstSpikeMin >  RstSpikeMax) && (rawTSMCcount >= RstSpikeMin)) return rawTSMCcount;
    }

    return 0;
}

/*}}}  */

/* Calibration file functions */
/*{{{  int ReadCalHeader (FILE *fp, int *version, long *valid_from)*/
int ReadCalHeader (FILE *fp, int *pversion, long *valid_from)
{
    int stat, year, month, day;
    char buf[128];

    /* Read the header info */
    stat = fscanf (fp, "BEGIN STAFF-FFT TC1 %*d-%*d-%*d %d %d-%d-%d",
                   pversion, &year, &month, &day);

    /* Skip the comments */
    do {
        fgets (buf, sizeof buf, fp);
    } while (strncmp (buf, "#END", 4) != 0);

return stat >= 4;
}

/*}}}  */
/*{{{  int ReadCalData (FILE *fp, struct caldata *pcd)*/
int ReadCalData (FILE *fp, struct caldata *pcd)
{
    int i;
    char buf[128];

    for (i = 0; i < 27; i++) {

        buf[0] = 0;
        fgets (buf, sizeof buf, fp);
        if (sscanf (buf, "%lf %*lf %lf %lf %lf %lf",
                &(pcd->Sx), &(pcd->Syz.x), &(pcd->Syz.y),
                &(pcd->Dyz.x), &(pcd->Dyz.y)) < 5) return 0;
        pcd++;
    }
    return 1;
}
/*}}}  */
/*{{{  int ProcCal (char *cal_name)*/
int ProcCal (char *cal_name)
{
    int version;
    FILE *fp;
    long valid_date;

    fp = fopen (cal_name, "r");
    if (fp == NULL) {
        fprintf (stderr, "Can't open calibration file '%s'\n", cal_name);
        return -1;
    }

    if (!ReadCalHeader (fp, &version, &valid_date)) {
        fprintf (stderr, "Error reading calibration file header\n");
        return -1;
    }

    if (!ReadCalData (fp, CalibrationData)) {
        fprintf (stderr, "Error reading calibration data\n");
        return -1;
    }

    fclose (fp);

    printf ("Read calibration file %s, version %d\n", cal_name, version);
    return version;
}
/*}}}  */
/*{{{  int DefaultCals ()*/
int DefaultCals ()
{
    int i;
    struct caldata  *pcd;

    pcd = CalibrationData;
    for (i = 0; i < 27; i++) {
        pcd->Sx    = 1.0;
        pcd->Syz.x = 1.0;
        pcd->Syz.y = 0.0;
        pcd->Dyz.x = 0.0;
        pcd->Dyz.y = 0.0;
        pcd++;
    }
    return 1;
}
/*}}}  */

/* Spin phase functions */
/*{{{  int ProcCorrel (unsigned char *buf, double pkt_time)*/
int ProcCorrel (unsigned char *buf, double pkt_time)
{
/* Records the times of the SRP, estimated from correlator packet times
   minus PEACE/Correlator SRP offset.  */

    static double last_srp_time = 0.0, last_cor_time = 0.0;
    static int last_stepped_level = 0;
    int stepped_level;

    /* Don't use correlator for SRP if SC HK file provided. */
    if (SC_HK_file) return 0;

    /* If correlator output more than 1/spin, save pkt_time and use
       after stepped energy level changes. */
    stepped_level = buf[7] & 0x0f;
    if (CorrelBitRateIndex != 0) {
        if (stepped_level == last_stepped_level) {
            last_cor_time = pkt_time;
            return 0;
        }
    }
    else last_cor_time = pkt_time;
    last_stepped_level = stepped_level;

    /* Try to measure spin period */
    if (fabs (last_cor_time - last_srp_time - NominalSpinPeriod) < 0.5)
        MeasuredSpinPeriod = last_cor_time - last_srp_time;
    last_srp_time = last_cor_time;

    SunRefPulse = last_cor_time - CorrelSRPOffset * MeasuredSpinPeriod;

    if (Verbosity & vSRP) {
        printf ("ProcCorrel: SRP %s, period %6.4f\n",
                doubledate (SunRefPulse), MeasuredSpinPeriod);
    }

    return 1;
}
/*}}}  */
/*{{{  int GetNewSRP ()*/
int GetNewSRP ()
{
    int status;
    unsigned char SC_HK_buf[256];
    static double new_srp, next_srp, last_srp, scrap;

    /* If no SC HK file, return failure. */
    if (SC_HK_file == NULL) return 0;

    /* If we read 2 SRPs last time, use second one now. */
    if (next_srp > CCSDS_time_ext ("2000-01-01T00:00:00.000Z")) {
        SunRefPulse = next_srp;
        next_srp = -1.0;
    }

    /* Else read another SC HK record */
    else {
        status = read_lrec (SC_HK_file, sizeof SC_HK_buf,
                            fmDDS_TLM, SC_HK_buf, &scrap);
        if (status < 0) return 0;
        new_srp  = CCSDS_time_ext (SC_HK_buf + 157);
        next_srp = CCSDS_time_ext (SC_HK_buf + 182);
        if (Verbosity & vSRP) {
            /* Note: must not use doubledate twice in one function call! */
            printf ("GetNewSRP (): SRP1 %s, ", doubledate (new_srp));
            printf ("SRP2 %s\n", doubledate (next_srp));
        }
        if (new_srp > CCSDS_time_ext ("2000-01-01T00:00:00.000Z"))
                                                SunRefPulse = new_srp;
        else return 1;
    }

    /* If last two SRPs have roughly correct separation, measure period */
    if (fabs (SunRefPulse - last_srp - NominalSpinPeriod) < 0.5)
        MeasuredSpinPeriod = SunRefPulse - last_srp;
    last_srp = SunRefPulse;

    return 1;
}
/*}}}  */
/*{{{  int ValidateSpin (double pkt_time)*/
int ValidateSpin (double pkt_time)
{
/* Given time tag of current FFT data packet, attempts to determine time tag of
   a suitable Sun Reference Pulse and measured spin period for determination
   of spin phase. If successful returns 1, if no suitable SRP is available
   returns 0. If spacecraft HK file is provided, then we read SRPs as needed
   from that file. Otherwise SRPs are obtained from correlator data. */


    /* If using SC HK file and we are more than one second after
       saved SRP, try to get another one. */
    if (SC_HK_file) while (pkt_time > SunRefPulse + 1.0 && GetNewSRP ());

    /* If measured period is zero, then we have not yet read enough SRPs
       to determine period, so must return failure. */
    if (MeasuredSpinPeriod == 0.0) return ValidSpin = 0;

    /* If within 1.5 periods of saved SRP, that SRP is valid. */
    if (fabs (pkt_time - SunRefPulse) < MeasuredSpinPeriod * 1.5)
                                         return ValidSpin = 1;

    /* Otherwise return failure. */
    return ValidSpin = 0;
}
/*}}}  */

/*{{{  int ProcFFT (unsigned char *buf, int pkt_size, double pkt_time)*/
int ProcFFT (unsigned char *buf, int pkt_size, double pkt_time)
{
    /*{{{  local variables*/
    /* Table giving raw acquisition interval in milliseconds for each
       combination of the disable x,y,z bits in the mode word. */
    static int RawAcqIntTab[] = { 200, 143, 143, 67, 143, 67, 67, 333 };
        /* Table giving packet sizes for modes 1 (or 3) in bytes for each
       combination of the disable x,y,z bits in the mode word. */
    static int PacketSizeTab[] = { 168,  88,  88,  34,  88,  34,  34, 0 };

	static double Average60HzPower = 0.0;
    
    struct tm *inttime;
    long ltime;
    int i, iphase;
    int rawacqint;     /* Raw acquisition interval in milli-seconds */
    short header[24];
    unsigned char swapbuf[256];
    struct specmat l2data[27];
    int coords;        /* Final co-ordinate system */
    double rotatex;    /* Rotation to be applied in radians */
    double spinphase;  /* Average spinphase as a fraction of a spin */
    double cen_time;   /* Average time in seconds past epoch */
	double power;
    float outdata[9];  /* Buffer for output data */
    /*}}}  */

    /* Check for emergency and high-res modes, which we can't handle */
    if (udefbits (buf) == 0 || udefbits (buf) == 2) return 0;

    /* Check packet size for low resolution data */
    if (pkt_size != PacketSizeTab[disableXYZ (buf)]) return 0;

    /* Copy first 6 bytes as is */
    memcpy (swapbuf, buf, 6);

    /* Copy other bytes, swapping adjacent pairs */
    memset (swapbuf + 6, sizeof swapbuf - 6, 0);
    _swab (buf + 6, swapbuf + 6, pkt_size - 6);

    ValidateSpin (pkt_time);

    /* Get the raw acquisition interval which depends
       on whether any components are disabled*/
    rawacqint = RawAcqIntTab[disableXYZ (swapbuf)];

    /* Skip packets subject to de-spin bug or Sun Pulse spike */
    if (udefbits(swapbuf) == 1 && (iphase = DespinBug (swapbuf, pkt_time)) != 0) {
        if (Verbosity & vSKIP)
            printf ("ProcFFT: %s de-spin bug or sun pulse spike skipped, phase %d\n",
                    doubledate (pkt_time), iphase);
		SkipCount++;
        return 0;
    }

    /* Skip packets subject to Reset Pulse spike */
    if (udefbits(swapbuf) == 1 && (iphase = ResetSpike (swapbuf)) != 0) {
        if (Verbosity & vSKIP)
            printf ("ProcFFT: %s reset spike skipped, MC count %d\n",
                    doubledate (pkt_time), iphase);
		SkipCount++;
        return 0;
    }

    /* Skip packets with mixed gains if calibrated */
    if ((CalVersion != 0) && higain (swapbuf) && logain (swapbuf)) {
        if (Verbosity & vGAIN)
            printf ("ProcFFT: %s mixed gain skipped\n", doubledate (pkt_time));
		SkipCount++;
        return 0;
    }

    /*{{{  Calculate time and spin phase at centre of data interval*/
    /* Note that this assumes that all raw time series contributing to this
       packet were acquired at the nominal intervals. This will not always be
       true for data acquired when the Reset Spike patch to the onboard software
       is active. */
    cen_time = pkt_time - (fftcount (swapbuf) - 1) * rawacqint / 2000.0 - 0.025;
    if (ValidSpin) spinphase = (cen_time - SunRefPulse) / MeasuredSpinPeriod;
    else spinphase = 0.0;
    
    /*}}}  */

    /* Decompress the TM data and convert to physical units */
    DecompFFT (swapbuf, pkt_time, l2data);

	/* Update average 60Hz power and skip packets that exceed average by specified limit */
	if (l2data[2].Py + l2data[2].Pz > 0.0) {
		power = 10.0 * log10 (l2data[2].Py + l2data[2].Pz);
		if (Average60HzPower == 0.0) Average60HzPower = power;
		else Average60HzPower = Average60HzPower * 0.9 + power * 0.1;
		power -= Average60HzPower;
		if (power > PowerLimit) {
			if (Verbosity & vSKIP)
		        printf ("ProcFFT: %s power limit spike skipped, power %3.1f\n",
                    doubledate (pkt_time), power);
			SkipCount++;
			return 0;
		}
	}

	/* If we get this far, increment the process count */
	OutputCount++;

    /*{{{  Decide what rotation, if any, to apply to spectral matrix.*/
    if (despinoff (swapbuf)) {
        if (ValidSpin && TransformControl == 1) {
            rotatex = -1.0 * TWOPI * spinphase;
            coords = 1;
        }
        else if (ValidSpin && TransformControl == 2) {
            rotatex = TWOPI * (SR2Rotation / 360.0 - spinphase);
            coords = 2;
        }
        else if (TransformControl == -3 || TransformControl == 3) {
            rotatex = TWOPI * SR2Rotation / 360.0;
            coords = 3;
        }
        else {
            rotatex = 0.0;
            coords = 0;
        }
    }
    else {
        if (ValidSpin && TransformControl == 0) {
            rotatex = TWOPI * spinphase;
            coords = 0;
        }
        else if (TransformControl == -2 || TransformControl == 2) {
            rotatex = TWOPI * SR2Rotation / 360.0;
            coords = 2;
        }
        else if (ValidSpin && TransformControl == 3) {
            rotatex = TWOPI * (spinphase + SR2Rotation / 360.0);
            coords = 3;
        }
        else {
            rotatex = 0.0;
            coords = 1;
        }
    }
    /*}}}  */

    /* Apply the resulting X rotation to all 27 frequency channels */
    if (rotatex != 0.0) for (i = 0; i < 27; i++) RotateX (l2data + i, rotatex);

    /*{{{  write the data to file*/
    if (Out_file != NULL) {
        /* Zero all words of header to start */
        for (i = 0; i < 24; i++) header[i] = 0;
    
        /* Level 2 file tag */
        header[0] = 'L' * 256 + '2';
    
        /* FFTL2 version number */
        header[1] = VERSION;
    
		/* Ground station ID */
		header[3] = GroundStation;

        /* Source instrument, FFT = 1 */
        header[4] = 1;
    
		/* TED diagnostics word */
		header[5] = TEDDiagnostic;

        /* Size of data field in bytes */
        header[6] = 9 * 27 * sizeof (float);
    
        /* Put in the timestamp - adjusted to refer to centre of interval */
        ltime = (long)cen_time;
        inttime = gmtime (&ltime);
        if (inttime != NULL) {
          header[7] = inttime->tm_year;
          header[8] = inttime->tm_mon+1;
          header[9] = inttime->tm_mday;
          header[10] = inttime->tm_hour;
          header[11] = inttime->tm_min;
          header[12] = inttime->tm_sec;
          header[13] = (int)((cen_time - floor (cen_time)) * 1000.0);
          }
    
        header[14] = sizeof (float);
    
        /* Put in the FFT mode */
        header[16] = 1;
    
        /* Time interval covered by spectrum */
        header[17] = fftcount (swapbuf) * rawacqint;
    
        /* Spin phase at centre of interval */
        if (ValidSpin) header[18] = (int)(spinphase * 4096.0) & 4095;
        else header[18] = -1;
    
        /* Output co-ordinate system */
        header[19] = coords;
    
        /* Calibration version */
        header[20] = CalVersion;
    
        /* FFT mode word */
        header[21] = (swapbuf[4] << 8) + swapbuf[5];
    
        /* Number of specified elements of the spectral maxtrix */
        header[22] = 9;
    
        /* Number of frequency channels */
        header[23] = 27;
    
        fwrite (header, sizeof header, 1, Out_file);
        if (coords == 2) for (i = 0; i < 27; i++) {
            /* For v1.03 onwards, for SR2 only
               we permute the axes so O3 (z) is spin azis */
            outdata[0] = l2data[i].Py;
            outdata[1] = l2data[i].Pz;
            outdata[2] = l2data[i].Px;
            outdata[3] = l2data[i].Cyz.x;
            outdata[4] = l2data[i].Cyz.y;
            outdata[5] = l2data[i].Cxz.x;
            outdata[6] = -l2data[i].Cxz.y;
            outdata[7] = l2data[i].Cxy.x;
            outdata[8] = -l2data[i].Cxy.y;
            fwrite (outdata, sizeof outdata, 1, Out_file);
        }
        else for (i = 0; i < 27; i++) {
            outdata[0] = l2data[i].Px;
            outdata[1] = l2data[i].Py;
            outdata[2] = l2data[i].Pz;
            outdata[3] = l2data[i].Cxy.x;
            outdata[4] = l2data[i].Cxy.y;
            outdata[5] = l2data[i].Cyz.x;
            outdata[6] = l2data[i].Cyz.y;
            outdata[7] = l2data[i].Cxz.x;
            outdata[8] = l2data[i].Cxz.y;
            fwrite (outdata, sizeof outdata, 1, Out_file);
        }
    }
    /*}}}  */
    if (Verbosity & vTIME) {
        printf ("ProcFFT: Pkt time %s,", doubledate (pkt_time));
        printf (" MC count %4d, VS %1d, spinphase %7.4lf, FFT count %2d\n",
                samp_clks (swapbuf), ValidSpin, spinphase, fftcount (swapbuf));
    }

    if (Verbosity & vDATA) ListData (cen_time, coords, l2data);


    return 1;
}
/*}}}  */
/*{{{  int ProcDSD (FILE *fp)*/
int ProcDSD (FILE *fp)
{
    unsigned char dsd_buf[MAXASMBSIZE];
    double pkt_time;
    int stat;
    int timeofday;

    while ((stat = ReadDSDPkt (fp, MAXASMBSIZE, dsd_buf, &pkt_time)) > 0) {
        if (pkt_time == -1.0)
           {
           /* no timestamp block producted by TED */
           if (Verbosity & vTIME)
             {
              printf ("ProcDSD: Pkt time undefined  ");
              printf ("DWP Master Clock 0x%04X, UDEF0 %1d  UDEF1 %1d  FFT Status Word 0x%04X\n",
                       (dsd_buf[2]*0x100+dsd_buf[3]*0x1),
                       ((dsd_buf[0] >> 1) & 1),
                       ((dsd_buf[0] >> 2) & 1),
                       (dsd_buf[4]*0x100+dsd_buf[5]*0x1));
             }
           }
        else {
            timeofday = (long)pkt_time % 86400;
            if (timeofday >= StartTime) ProcFFT (dsd_buf, stat, pkt_time);
            if (timeofday > EndTime) return 0;
        }
    }
    return stat;
}
/*}}}  */

/* Call back functions for 'tlmproc' module */
/*{{{  int proc_hk (hk_ptr, pkt_size, ob_time)*/
int proc_hk (hk_ptr, pkt_size, ob_time)
unsigned char *hk_ptr;
int pkt_size;
double ob_time;
{
CorrelSRPOffset = peace_srp (hk_ptr) / 4096.0;
CorrelBitRateIndex = corbitrate (hk_ptr);

if (Verbosity & vHK) {
    printf ("proc_hk: time %s, EW5RSCNT %d\n",
            doubledate (ob_time), hk_rst_cnt (hk_ptr));
}

return 1;
}
/*}}}  */
/*{{{  int proc_raw (hk_ptr, pkt_size, ob_time)*/
int proc_raw (hk_ptr, pkt_size, ob_time)
unsigned char *hk_ptr;
int pkt_size;
double ob_time;
{
return 1;
}
/*}}}  */
/*{{{  int proc_unused (hk_ptr, pkt_size)*/
int proc_unused (ptr, pkt_size, ob_time)
unsigned char *ptr;
int pkt_size;
double ob_time;
{
return 1;
}
/*}}}  */
/*{{{  int proc_sci (tlm_ptr, pkt_size, ob_time, ss_offset)*/
int proc_sci (tlm_ptr, pkt_size, ob_time, ss_offset)
unsigned char *tlm_ptr;
int pkt_size;
double ob_time, ss_offset;
{
double pkt_time;
int src;

src = source(tlm_ptr);
pkt_time = ob_time + samp_clks (tlm_ptr) / 900.0;

if (src == 1) ProcFFT (tlm_ptr, pkt_size, pkt_time);
if (src == 5) ProcCorrel (tlm_ptr, pkt_time);

return 1;
}
/*}}}  */
/*{{{  void write_err (pkt_time, message)*/
void write_err (pkt_time, message)
double pkt_time;
char *message;
{
fprintf (stdout, "Error: (%s) %s\n", doubledate (pkt_time), message);
}
/*}}}  */

/*{{{  void usage ()*/
void usage () {
    printf ("usage: fftl2 [options]\n"
            "       -h staff_hk_file\n"
            "       -p spacecraft_hk_file (for Sun Reference Pulse times)\n"
            "       -n staff_sci_file\n"
            "       -d dsd_file (excludes -h and -n)\n"
            "       -c calibration_file\n"
            "       -o output_file\n"
            "       -v verbosity\n"
            "       -r rotation from de-spun ant to SR2 (degrees)\n"
            "       -t transform to co-ordinates (0=ant, 1=despun ant, 2=SR2)\n"
            "       -s start_time (hh:mm:ss)\n"
            "       -e end_time (hh:mm:ss)\n"
            "       -a min:max (MC count limits for reset spike)\n"
            "       -b min:max (phase limits for de-spin bug or SRP spike)\n"
			"       -l limit (suppress spikes with power above limit)\n");
}
/*}}}  */
/*{{{  int procarg (int argc, char** argv)*/
int procarg (int argc, char** argv)
{
    int n;

    n = 0;
    while (argc > 0) {
        /* All options of the form "-%c", where %c is one character. */
        if (argv[0][0] != '-' || argv[0][2] != 0) {
            fprintf (stderr, "Don't understand option %s\n", *argv);
            return -1;
        }
        switch (argv[0][1]) {
            case 'h': STA_HK_name = argv[1];
            break;
            case 'n': STA_SCI_name = argv[1];
            break;
            case 'd': DSD_name = argv[1];
            break;
            case 'p': SC_HK_name = argv[1];
            break;
            case 'c': Cal_name = argv[1];
            break;
            case 'o': Out_name = argv[1];
            break;
            case 'v': Verbosity = atoi (argv[1]);
            break;
            case 'r': SR2Rotation = atof (argv[1]);
            break;
            case 't': TransformControl = atoi (argv[1]);
            break;
            case 's': StartTime = encode_time (argv[1]);
            break;
            case 'e': EndTime = encode_time (argv[1]);
            break;
            case 'b': sscanf (argv[1], "%d:%d", &BugPhaseMin, &BugPhaseMax);
            break;
            case 'a': sscanf (argv[1], "%d:%d", &RstSpikeMin, &RstSpikeMax);
			break;
			case 'l': PowerLimit = atof (argv[1]);
            break;
            default:
            fprintf (stderr, "Don't understand option %s\n", argv[0]);
            return -1;
        }
        argc -= 2;
        argv += 2;
        n++;
    }
    return n;
}
/*}}}  */

/*{{{  main*/
void main (int argc, char **argv)
{
    int stat;

    printf ("FFTL2 version %5.2lf\n", (double)VERSION / 100.0);

    if (procarg (argc - 1, argv + 1) < 1) {
        usage ();
        exit (0);
    }

    /* Set the timezone to GMT */
#ifdef _MSC_VER
    _putenv ("TZ=GMT");
#else
    putenv ("TZ=GMT");
#endif
    tzset ();

    /*{{{  open SC HK file*/
    if (SC_HK_name) {
        SC_HK_file = fopen (SC_HK_name, "rb");
        if (SC_HK_file == NULL) {
            fprintf (stderr, "Can't open SC HK file '%s'\n", SC_HK_name);
            exit (1);
        }
    }
    /*}}}  */
    /*{{{  process calibration file*/
    DefaultCals ();
    if (Cal_name) {
        CalVersion = ProcCal (Cal_name);
        if (CalVersion < 0) {
            fprintf (stderr, "Error processing calibration file '%s'\n", Cal_name);
            exit (1);
        }
    }
    /*}}}  */
    /*{{{  open output file*/
    if (Out_name) {
        Out_file = fopen (Out_name, "wb");
        if (Out_file == NULL) {
            fprintf (stderr, "Can't open output file '%s'\n", Out_name);
            exit (1);
        }
    }
    /*}}}  */
    /*{{{  open and process DSD file*/
    if (DSD_name) {
        if (STA_HK_name || STA_SCI_name) {
            fprintf (stderr, "presence of -d excludes -h and -n\n");
            exit (1);
        }
        DSD_file = fopen (DSD_name, "rb");
        if (DSD_file == NULL) {
            fprintf (stderr, "Can't open DSD file '%s'\n", DSD_name);
            exit (1);
        }
        if ((stat = ProcDSD (DSD_file)) < -1) {
            fprintf (stderr, "Fatal error %d processing DSD file at offset %d.\n",
                              stat, ftell (DSD_file));
            exit (1);
        }
    }
    /*}}}  */
    /*{{{  open and process HK/SCI files*/
    if (!DSD_name) {
        if (!STA_HK_name || !STA_SCI_name) {
            fprintf (stderr, "absence of -d requires -h and -n\n");
            exit (1);
        }
        HK_file = fopen (STA_HK_name, "rb");
        if (HK_file == NULL) {
            fprintf (stderr, "Can't open HK data file '%s'\n", STA_HK_name);
            exit (1);
        }
        SCI_file = fopen (STA_SCI_name, "rb");
        if (SCI_file == NULL) {
            fprintf (stderr, "Can't open SCI data file '%s'\n", STA_SCI_name);
            exit (1);
        }
        if (proc_file (HK_file, SCI_file, StartTime, EndTime, fmDDS_TLM) == 0) {
            fprintf (stderr, "Fatal error processing HK/SCI files.\n");
            exit (1);
        }
    }
    /*}}}  */

    /*{{{  close any open files*/
    if (SC_HK_file != NULL) fclose (SC_HK_file);
    if (HK_file != NULL) fclose (HK_file);
    if (SCI_file != NULL) fclose (SCI_file);
    if (DSD_file != NULL) fclose (DSD_file);
    if (Out_file != NULL) fclose (Out_file);
    /*}}}  */

	printf ("Finished processing, %d packets output, %d skipped.\n", OutputCount, SkipCount);
    exit (0);
}
/*}}}  */

